// SPDX-FileCopyrightText: 2025 Moritz Hanke <hanke@dkrz.de>
//
// SPDX-License-Identifier: BSD-3-Clause

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <mpi.h>

#include "yac.h"

#include "elmer_coupling.h"
#include "elmer_icon_coupling.h"
#include "elmer_grid.h"

static int comp_id = -1;
static int comm_rank = -1, comm_size = -1;

static void compute_cell_centers(
  int nbr_cells, int * cell_to_vertex, int * num_vertices_per_cell,
  double * x_vertices, double * y_vertices,
  double * x_cells, double * y_cells);

void coupling_init(char const * coupling_config_file, MPI_Comm * elmer_comm) {

  // initialise YAC
  // * is collective operation on MPI_COMM_WORLD
  // * should be called as early as possible
  // * in case not all Elmer processes want to take part in the coupling
  //   this has to be modified
  //   (see:
  //     https://dkrz-sw.gitlab-pages.dkrz.de/yac/d4/d40/init_yac_detail.html)
  // * will call MPI_Init, if not yet called by the user
  yac_cinit();

  // read configuration file
  // * contains calendar, start- and end-date
  // * defines couplings
  yac_cread_config_yaml(coupling_config_file);

  // define component
  // * is collective operation for all processes that initialised YAC
  yac_cdef_comp(ELMER_COMP_NAME, &comp_id);

  // get communicator for all processes that are part of the Elmer component
  // * this is a collective operation for all processes in the component
  // * Elmer should only use this communicator internally and access
  //   MPI_COMM_WORLD at all
  yac_cget_comp_comm(comp_id, elmer_comm);

  // get number of ranks for the elmer component
  // (required for reading in the grid data)
  MPI_Comm_rank(*elmer_comm, &comm_rank);
  MPI_Comm_size(*elmer_comm, &comm_size);
}

void coupling_setup(char const * grid_dir, int num_parts) {

  // get grid data from elmer component
  // in the case of the dummy, we have to read it from file

  // read grid data from file
  // * each process only reads in its local part of the grid
  // * in Elmer this information probably already available and does not have
  //   to be read from file
  int nbr_vertices;
  int nbr_cells;
  int * num_vertices_per_cell;
  double * x_vertices;
  double * y_vertices;
  int * cell_ids;
  int * vertex_ids;
  int * cell_to_vertex;
  read_grid(
    grid_dir, comm_rank, comm_size, num_parts,
    &nbr_vertices, &nbr_cells, &num_vertices_per_cell,
    &x_vertices, &y_vertices, &cell_ids, &vertex_ids, &cell_to_vertex);

  // register Elmer grid in YAC
  // * is defined as an unstructured grid
  int grid_id;
  yac_cdef_grid_unstruct(
    ELMER_GRID_NAME, nbr_vertices, nbr_cells, num_vertices_per_cell,
    x_vertices, y_vertices, cell_to_vertex, &grid_id);

  // define global ids for all cells and vertices
  // * this information is used to identify cells/vertice between processes
  // * if this information is not provided, YAC will generate it (resulting
  //   IDs then depend on the number of processes)
  yac_cset_global_index(cell_ids, YAC_LOCATION_CELL, grid_id);
  yac_cset_global_index(vertex_ids, YAC_LOCATION_CORNER, grid_id);

  double * x_cells = malloc(nbr_cells * sizeof(*x_cells));
  double * y_cells = malloc(nbr_cells * sizeof(*y_cells));
  compute_cell_centers(
    nbr_cells, cell_to_vertex, num_vertices_per_cell,
    x_vertices, y_vertices, x_cells, y_cells);

  // define location at which field data will be provided
  // * an arbitrary number of locations can be defined (e.g. at vertices,
  //   cell centers, edge middle points)
  int corner_point_id, cell_point_id;
  yac_cdef_points_unstruct(
    grid_id, nbr_vertices, YAC_LOCATION_CORNER, x_vertices, y_vertices,
    &corner_point_id);
  yac_cdef_points_unstruct(
    grid_id, nbr_cells, YAC_LOCATION_CELL, x_cells, y_cells, &cell_point_id);

  // cleanup of grid data
  // * YAC has internal copy
  free(y_cells);
  free(x_cells);
  free(num_vertices_per_cell);
  free(vertex_ids);
  free(cell_ids);
  free(x_vertices);
  free(y_vertices);
  free(cell_to_vertex);

  // construct coupling between Elmer/Ice and ICON
  construct_elmer_icon_coupling(comp_id, corner_point_id, cell_point_id);

  // sychronizes all definitions between all components
  // * afterwards the exchange information can be queried
  // * this is optional
  yac_csync_def();

  // construct coupling between Elmer/Ice and ICON (using sychronized
  // information from all components)
  construct_elmer_icon_coupling_post_sync(comm_rank);

  // end of definition phase
  // * collective operation for all processes that initialised YAC
  // * exchanges information between all processes (various query routines are
  //   available to access this data; also from other components)
  // * computes weights required for all defined couples
  yac_cenddef();
}

void coupling_finalize() {

  // finalise YAC
  // * if user has called MPI_Init, he also has to call MPI_Finalize afterwards
  yac_cfinalize();
}

static inline void LLtoXYZ(double lon, double lat, double p_out[]) {

   while (lon < -M_PI) lon += 2.0 * M_PI;
   while (lon >= M_PI) lon -= 2.0 * M_PI;

   double cos_lat = cos(lat);
   p_out[0] = cos_lat * cos(lon);
   p_out[1] = cos_lat * sin(lon);
   p_out[2] = sin(lat);
}

static inline void XYZtoLL (double const p_in[], double * lon, double * lat) {

   *lon = atan2(p_in[1] , p_in[0]);
   *lat = M_PI_2 - acos(p_in[2]);
}

static inline void normalise_vector(double v[]) {

   double norm = 1.0 / sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);

   v[0] *= norm;
   v[1] *= norm;
   v[2] *= norm;
}

static void compute_cell_centers(
  int nbr_cells, int * cell_to_vertex, int * num_vertices_per_cell,
  double * x_vertices, double * y_vertices,
  double * x_cells, double * y_cells) {

  for (int i = 0; i < nbr_cells; ++i) {

    int num_vertices = num_vertices_per_cell[i];
    double cell_center[3] = {0.0, 0.0, 0.0};

    for (int j = 0; j < num_vertices; ++j) {
      double vertex[3];
      LLtoXYZ(
        x_vertices[cell_to_vertex[j]], y_vertices[cell_to_vertex[j]], vertex);
      cell_center[0] += vertex[0];
      cell_center[1] += vertex[1];
      cell_center[2] += vertex[2];
    }

    normalise_vector(cell_center);
    XYZtoLL(cell_center, x_cells + i, y_cells + i);

    cell_to_vertex += num_vertices;
  }
}
