// SPDX-FileCopyrightText: 2025 Moritz Hanke <hanke@dkrz.de>
//
// SPDX-License-Identifier: BSD-3-Clause

#include <stdlib.h>
#include <stdio.h>

#include "yac.h"

#include "elmer_coupling.h"
#include "elmer_icon_coupling.h"

static int clt_field_id = -1;
static char const clt_field_name[] = "clt";
static int clt_collection_size = 1;
char const clt_field_timestep[] = "P1M";
static double * clt_field = NULL;

static int pr_field_id = -1;
static char const pr_field_name[] = "pr";
static int pr_collection_size = 1;
char const pr_field_timestep[] = "P1M";
static double * pr_field = NULL;

static char const * field_names[] = {clt_field_name, pr_field_name};
enum {NUM_FIELDS = sizeof(field_names) / sizeof(field_names[0])};

void construct_elmer_icon_coupling(
  int comp_id, int corner_point_id, int cell_point_id) {

  size_t nbr_vertices = yac_cget_points_size(corner_point_id);
  size_t nbr_cells = yac_cget_points_size(cell_point_id);

  // register total cloud cover field in YAC
  yac_cdef_field(
    clt_field_name, comp_id, &corner_point_id, 1, clt_collection_size,
    clt_field_timestep, YAC_TIME_UNIT_ISO_FORMAT, &clt_field_id);

  // allocate and initialise total cloud cover field buffer
  clt_field = malloc(nbr_vertices * sizeof(*clt_field));
  for (int i = 0; i < nbr_vertices; ++i) clt_field[i] = 0.0;

  // register precipitation flux field in YAC
  yac_cdef_field(
    pr_field_name, comp_id, &cell_point_id, 1, pr_collection_size,
    pr_field_timestep, YAC_TIME_UNIT_ISO_FORMAT, &pr_field_id);

  // allocate and initialise precipitation flux field buffer
  pr_field = malloc(nbr_cells * sizeof(*pr_field));
  for (int i = 0; i < nbr_cells; ++i) pr_field[i] = 0.0;
}

void construct_elmer_icon_coupling_post_sync(int comm_rank) {

  // after synchronisation or the end of the definition phase YAC can be
  // queried about various information

  for (size_t i = 0; i < NUM_FIELDS; ++i) {

    char const * field_name = field_names[i];
    if (yac_cget_field_role(
          ELMER_COMP_NAME, ELMER_GRID_NAME, field_name) ==
          YAC_EXCHANGE_TYPE_TARGET) {

      char const * src_comp_name;
      char const * src_grid_name;
      char const * src_field_name;
      yac_cget_field_source(
        ELMER_COMP_NAME, ELMER_GRID_NAME, field_name,
        &src_comp_name, &src_grid_name, &src_field_name);
      char const * src_field_timestep =
        yac_cget_field_timestep(src_comp_name, src_grid_name, src_field_name);
      char const * src_field_metadata =
        yac_cget_field_metadata(src_comp_name, src_grid_name, src_field_name);

      if (comm_rank == 0)
        fprintf(
          stderr,
          "field \"%s\":\n"
          " - source:\n"
          "   - component: %s\n"
          "   - grid:      %s\n"
          "   - timestep:  %s\n"
          "   - metadata:  %s\n",
          field_name, src_comp_name, src_grid_name,
          src_field_timestep, src_field_metadata);
    }
  }
}

void elmer_icon_coupling(int comm_rank) {

  // checks whether the total cloud cover field is defined as a target
  // in a couple
  if (yac_cget_role_from_field_id(clt_field_id) ==
      YAC_EXCHANGE_TYPE_TARGET) {

    int info, err;

    if (comm_rank == 0) {

      // get the action executed by YAC in the next get operation called for
      // the total cloud cover field and print out some information
      yac_cget_action(clt_field_id, &info);
      fprintf(
        stderr,
        "call get for field: \"%s\" datatime: %s action: %s\n",
        clt_field_name, yac_cget_field_datetime(clt_field_id),
        ((info == YAC_ACTION_COUPLING) ||
         (info == YAC_ACTION_GET_FOR_RESTART))?"couping":"none");
    }

    // execute get operation for total cloud cover field
    // * if this is a coupling timestep, this will block until the data has
    //   been received
    // * if this is not a coupling timestep, total cloud cover field buffer
    //   is left untouched and routine will return immediately
    yac_cget_(
      clt_field_id, clt_collection_size, clt_field, &info, &err);

    // if this was a coupling timestep
    if ((info == YAC_ACTION_COUPLING) ||
        (info == YAC_ACTION_GET_FOR_RESTART)) {

      // prepare received data for elmer

      // update elmer internal total cloud cover field

    }
  }

  // checks whether the precipitation flux field is defined as a target
  // in a couple
  if (yac_cget_role_from_field_id(pr_field_id) ==
      YAC_EXCHANGE_TYPE_TARGET) {

    int info, err;

    if (comm_rank == 0) {

      // get the action executed by YAC in the next get operation called for
      // the precipitation flux field and print out some information
      yac_cget_action(pr_field_id, &info);
      fprintf(
        stderr,
        "call get for field: \"%s\" datatime: %s action: %s\n",
        pr_field_name, yac_cget_field_datetime(pr_field_id),
        ((info == YAC_ACTION_COUPLING) ||
         (info == YAC_ACTION_GET_FOR_RESTART))?"couping":"none");
    }

    // execute get operation for precipitation flux field
    // * if this is a coupling timestep, this will block until the data has
    //   been received
    // * if this is not a coupling timestep, precipitation flux field buffer
    //   is left untouched and routine will return immediately
    yac_cget_(
      pr_field_id, pr_collection_size, pr_field, &info, &err);

    // if this was a coupling timestep
    if ((info == YAC_ACTION_COUPLING) ||
        (info == YAC_ACTION_GET_FOR_RESTART)) {

      // prepare received data for elmer

      // update elmer internal precipitation flux field

    }
  }
}

void destruct_elmer_icon_coupling() {

  free(pr_field);
  free(clt_field);
}
