// SPDX-FileCopyrightText: 2025 Moritz Hanke <hanke@dkrz.de>
//
// SPDX-License-Identifier: BSD-3-Clause

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <netcdf.h>

#include "yac.h"
#include "yac_utils.h"

#include "icon_varlist.h"

int main (int argc, char *argv[]) {

  // initialise YAC
  // * is collective operation on MPI_COMM_WORLD
  // * should be called as early as possible
  // * in case not all Elmer processes want to take part in the coupling
  //   this has to be modified
  //   (see:
  //     https://dkrz-sw.gitlab-pages.dkrz.de/yac/d4/d40/init_yac_detail.html)
  // * will call MPI_Init, if not yet called by the user
  yac_cinit();

  // define component
  // * is collective operation for all processes that initialised YAC
  int comp_id;
  yac_cdef_comp("icon", &comp_id);

  // get communicator for all processes that are part of the Elmer component
  // * this is a collective operation for all processes in the component
  // * Elmer should only use this communicator internally and access
  //   MPI_COMM_WORLD at all
  MPI_Comm icon_comm;
  yac_cget_comp_comm(comp_id, &icon_comm);

  int comm_rank, comm_size;
  MPI_Comm_rank(icon_comm, &comm_rank);
  MPI_Comm_size(icon_comm, &comm_size);

  if (comm_size != 1) {
    fprintf(stderr, "wrong number of process for %s (has to be 1)\n", argv[0]);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  if (argc != 4) {
    if (comm_rank == 0)
      fprintf(
        stderr,
        "wrong number of arguments\n"
        "usage: %s grid_file data_file varlist_file\n"
        "  - grid_file: path to icon grid file\n"
        "  - data_file: path to icon data file\n"
        "  - varlist_file: path to file conaining \n"
        "    list of variables in data_file\n", argv[0]);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  char const * grid_filename = argv[1];
  char const * data_filename = argv[2];
  char const * varlist_filename = argv[3];

  // read icon grid data from file
  int num_vertices;
  int num_cells;
  int * num_vertices_per_cell;
  int * cell_to_vertex;
  double * x_vertices;
  double * y_vertices;
  double * x_cells;
  double * y_cells;
  int * cell_mask;
  yac_read_icon_grid_information(
    grid_filename, &num_vertices, &num_cells, &num_vertices_per_cell,
    &cell_to_vertex, &x_vertices, &y_vertices, &x_cells, &y_cells, &cell_mask);

  // register ICON grid in YAC
  // * is defined as an unstructured grid
  int grid_id;
  yac_cdef_grid_unstruct(
    "icon_grid", num_vertices, num_cells, num_vertices_per_cell,
    x_vertices, y_vertices, cell_to_vertex, &grid_id);

  // define global ids for all cells and vertices
  // * this information is used to identify cells/vertice between processes
  // * if this information is not provided, YAC will generate it (resulting
  //   IDs then depend on the number of processes)
  int * cell_ids = malloc(num_cells * sizeof(*cell_ids));
  int * vertex_ids = malloc(num_vertices * sizeof(*vertex_ids));
  for (int i = 0; i < num_cells; ++i) cell_ids[i] = i;
  for (int i = 0; i < num_vertices; ++i) vertex_ids[i] = i;
  yac_cset_global_index(cell_ids, YAC_LOCATION_CELL, grid_id);
  yac_cset_global_index(vertex_ids, YAC_LOCATION_CORNER, grid_id);
  free(vertex_ids);
  free(cell_ids);

  // define location at which field data will be provided
  // * an arbitrary number of locations can be defined (e.g. at vertices,
  //   cell centers, edge middle points)
  int cell_point_id;
  yac_cdef_points_unstruct(
    grid_id, num_cells, YAC_LOCATION_CELL, x_cells, y_cells,
    &cell_point_id);

  // cleanup of grid data
  // * YAC has internal copy
  free(num_vertices_per_cell);
  free(cell_to_vertex);
  free(x_vertices);
  free(y_vertices);
  free(x_cells);
  free(y_cells);
  free(cell_mask);

  // get list of all variables
  struct icon_varlist * varlist;
  get_icon_varlist(varlist_filename, data_filename, &varlist);

  if (varlist->ncells != num_cells) {
    fputs("number of cells in grid and data file do not match\n", stderr);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  // define fields available in the data file
  int collection_size = 1;
  char const * data_timestep = "P1M";
  int * field_ids = malloc(varlist->num_vars * sizeof(*field_ids));
  yac_cdef_calendar(YAC_PROLEPTIC_GREGORIAN);
  for (size_t i = 0; i < varlist->num_vars; ++i) {
    yac_cdef_field(
      varlist->vars[i].name, comp_id, &cell_point_id, 1,
      collection_size, data_timestep, YAC_TIME_UNIT_ISO_FORMAT, field_ids + i);
    char field_metadata[128];
    sprintf(
      field_metadata, "long_name: %s ; units: %s",
      varlist->vars[i].long_name, varlist->vars[i].units);
    yac_cdef_field_metadata(
      "icon", "icon_grid", varlist->vars[i].name, field_metadata);
  }

  // end of definition phase
  // * collective operation for all processes that initialised YAC
  // * exchanges information between all processes (various query routines are
  //   available to access this data; also from other components)
  // * computes weights required for all defined couples
  yac_cenddef();

  // after the end of the definition phase we can query which fields were
  // actually requested by another component
  int * field_is_source = malloc(varlist->num_vars * sizeof(*field_is_source));
  for (int i = 0; i < varlist->num_vars; ++i)
    field_is_source[i] =
      yac_cget_role_from_field_id(field_ids[i]) == YAC_EXCHANGE_TYPE_SOURCE;

  double * field_buffer = malloc(varlist->ncells * sizeof(*field_buffer));

  // time stepping loop
  int const nstep = 12;
  for (int step = 0; step < nstep; ++step) {

    for (size_t i = 0; i < varlist->num_vars; ++i) {

      if (field_is_source[i]) {

        size_t start[2], count[2];
        start[0] = (size_t)step;
        start[1] = 0;
        count[0] = 1;
        count[1] = varlist->ncells;
        YAC_HANDLE_ERROR(
          nc_get_vara_double(
            varlist->ncid, varlist->vars[i].varid, start, count, field_buffer));

        int info, err;
        yac_cput_(field_ids[i], collection_size, field_buffer, &info, &err);
      }
    }
  }

  free(field_buffer);
  free(field_is_source);
  free(field_ids);
  free_icon_varlist(varlist);

  MPI_Comm_free(&icon_comm);

  // finalise YAC
  // * if user has called MPI_Init, he also has to call MPI_Finalize afterwards
  yac_cfinalize();

  return EXIT_SUCCESS;
}
